@@ -1,21 +1,21 @@
-from typing import List, Tuple, Union
-from foundry_sdk.db_mgmt import SQLDatabase, InsertionMode
+
+from foundry_sdk.db_mgmt import InsertionMode, SQLDatabase
 
 
 def insert_data(
     db: SQLDatabase,
     table: str,
-    columns: Union[str, List[str]],
-    data: List[Tuple],
+    columns: str | list[str],
+    data: list[tuple],
     insertion_mode: InsertionMode,
     condition: str = None,
 ) -> None:
-    """Retrieve data from a table in the database.
+    """
+    Retrieve data from a table in the database.
 
     Ensures column names are enclosed in double quotes for PostgreSQL.
     Allows single column name as a string or multiple as a list.
     """
-
     # Enclose column names in double quotes
     columns_string = ", ".join(f'"{col}"' for col in columns)
 
