@@ -1,10 +1,10 @@
-from typing import List, Tuple, Union, Optional
+
 from foundry_sdk.db_mgmt import SQLDatabase
 
 
 def retrieve_data(
-    db: SQLDatabase, table: str, columns: Union[str, List[str]], condition: Optional[str] = None
-) -> List[Tuple]:
+    db: SQLDatabase, table: str, columns: str | list[str], condition: str | None = None
+) -> list[tuple]:
     """
     Retrieve data from a table in the database.
 
