@@ -1,7 +1,6 @@
 import io
 import logging
 import multiprocessing as mp
-import os
 import time
 import typing as t
 from contextlib import contextmanager
@@ -310,16 +309,14 @@ class SQLAlchemyDatabase:
         data: pl.DataFrame,
         mode: t.Literal["RAISE", "IGNORE", "UPDATE"],
         *,
-        threshold_for_copy: int = 1_000_000,  # temporarily do not use copy
+        threshold_for_copy: int = 1_000_001,
         returning_id: bool = False,
         parallel_workers: int = 1,
     ) -> pl.DataFrame | None:
-        
         """
         Validate a Polars DataFrame against the model schema, then delegate
         to the bulk-insert implementation.
         """
-
         self.check_mode(mode)
         # 1. Basic type & column-name checks
         if not isinstance(data, pl.DataFrame):
@@ -579,11 +576,11 @@ class SQLAlchemyDatabase:
                     ids.append(row_dict)
                 if hasattr(progress_bar, "update"):
                     progress_bar.update(len(key_rows))
-                
+
                 # Commit every 100 iterations to avoid hitting connection limits
                 if (i + 1) % 100 == 0:
                     session.commit()
-            
+
             # Final commit for any remaining uncommitted work
             session.commit()
         return ids
@@ -594,7 +591,7 @@ class SQLAlchemyDatabase:
         data: pl.DataFrame,
         mode: t.Literal["RAISE", "IGNORE", "UPDATE"],
         *,
-        chunk_size: int = 1_000_000_000,
+        chunk_size: int = 1_000_000,
         returning_id: bool = False,
     ) -> pl.DataFrame | None:
         """
@@ -632,7 +629,7 @@ class SQLAlchemyDatabase:
             )
 
         all_id_maps: list[pl.DataFrame] = []
-        
+
         # Performance tracking
         total_times = {"csv_gen": 0, "copy_data": 0, "insert": 0, "fetch_ids": 0, "truncate": 0}
         batch_count = 0
@@ -671,7 +668,7 @@ class SQLAlchemyDatabase:
                 batch = data.slice(offset, chunk_size)
                 if batch.height == 0:
                     continue
-                
+
                 batch_count += 1
 
                 # 1) Time truncate operation
@@ -729,13 +726,13 @@ class SQLAlchemyDatabase:
         total_rows = data.height
         print(
             f"COPY Performance Breakdown for {total_rows:,} rows in {batch_count} batches:\n"
-            f"  CSV generation: {total_times['csv_gen']:.3f}s ({total_times['csv_gen']/total_time*100:.1f}%)\n"
-            f"  COPY to temp:   {total_times['copy_data']:.3f}s ({total_times['copy_data']/total_time*100:.1f}%)\n"
-            f"  INSERT upsert:  {total_times['insert']:.3f}s ({total_times['insert']/total_time*100:.1f}%)\n"
-            f"  Fetch IDs:      {total_times['fetch_ids']:.3f}s ({total_times['fetch_ids']/total_time*100:.1f}%)\n"
-            f"  Truncate:       {total_times['truncate']:.3f}s ({total_times['truncate']/total_time*100:.1f}%)\n"
+            f"  CSV generation: {total_times['csv_gen']:.3f}s ({total_times['csv_gen'] / total_time * 100:.1f}%)\n"
+            f"  COPY to temp:   {total_times['copy_data']:.3f}s ({total_times['copy_data'] / total_time * 100:.1f}%)\n"
+            f"  INSERT upsert:  {total_times['insert']:.3f}s ({total_times['insert'] / total_time * 100:.1f}%)\n"
+            f"  Fetch IDs:      {total_times['fetch_ids']:.3f}s ({total_times['fetch_ids'] / total_time * 100:.1f}%)\n"
+            f"  Truncate:       {total_times['truncate']:.3f}s ({total_times['truncate'] / total_time * 100:.1f}%)\n"
             f"  Total time:     {total_time:.3f}s\n"
-            f"  Overall rate:   {total_rows/total_time:.0f} rows/sec"
+            f"  Overall rate:   {total_rows / total_time:.0f} rows/sec"
         )
 
         if returning_id:
@@ -792,8 +789,8 @@ class SQLAlchemyDatabase:
 
         # Test multiprocessing capability first
         try:
-            import multiprocessing
-            mp.set_start_method('spawn', force=True)  # Force spawn method for better compatibility
+
+            mp.set_start_method("spawn", force=True)  # Force spawn method for better compatibility
             logger.info(f"Multiprocessing start method set to: {mp.get_start_method()}")
         except Exception as e:
             logger.warning(f"Could not set multiprocessing start method: {e}")
@@ -848,7 +845,7 @@ class SQLAlchemyDatabase:
                     progress_bar.update(batch_size)
                 if returning_id and id_df is not None:
                     all_id_maps.append(id_df)
-        
+
         if show_progress:
             progress_bar.close()
 
@@ -935,9 +932,11 @@ def _copy_worker_static(
     returning_id: bool,
 ) -> tuple[int, pl.DataFrame | None]:
     """
-    Static worker function for parallel COPY operations.
+    Static worker function for parallel COPY operations with pre-filtering optimization.
     This function recreates its own database connection from the connection URL.
 
+    Pre-filters existing data to reduce conflict resolution overhead and improve performance.
+
     Returns:
         tuple: (batch_size, id_dataframe_or_none)
 
@@ -954,25 +953,14 @@ def _copy_worker_static(
         # Connection tuning for bulk operations
         cursor.execute("SET synchronous_commit = OFF")
 
-        # Precompute ON CONFLICT clause based on mode
-        conflict_cols = ", ".join(f'"{c}"' for c in pks)
-        if mode == "RAISE":
-            on_conflict = ""
-        elif mode == "IGNORE":
-            on_conflict = f"ON CONFLICT ({conflict_cols}) DO NOTHING"
-        else:  # UPDATE
-            non_pk_cols = [c for c in cols if c not in pks]
-            if non_pk_cols:
-                set_clause = ", ".join(f'"{c}" = EXCLUDED."{c}"' for c in non_pk_cols)
-                on_conflict = f"ON CONFLICT ({conflict_cols}) DO UPDATE SET {set_clause}"
-            else:
-                on_conflict = f"ON CONFLICT ({conflict_cols}) DO NOTHING"
-
-        # Create temp table for this worker
+        # Create temp tables for this worker
         temp_table = f"temp_{table}_worker_{offset}"
+        existing_keys_table = f"temp_existing_keys_worker_{offset}"
         non_id_cols = [c for c in cols if c.upper() != "ID"]
         non_id_col_list = ", ".join(f'"{c}"' for c in non_id_cols)
+        pk_col_list = ", ".join(f'"{c}"' for c in pks)
 
+        # Create staging table for new data
         cursor.execute(
             f'CREATE TEMP TABLE "{temp_table}" AS SELECT {non_id_col_list} FROM "{table}" WHERE FALSE'  # noqa: S608
         )
@@ -983,26 +971,67 @@ def _copy_worker_static(
         buf.seek(0)
         cursor.copy_expert(f'COPY "{temp_table}" ({non_id_col_list}) FROM STDIN WITH CSV HEADER', buf)
 
-        # 2) INSERT FROM staging into real table
-        insert_sql = (
-            f'INSERT INTO "{table}" ({non_id_col_list}) SELECT {non_id_col_list} FROM "{temp_table}" '  # noqa: S608
-            + on_conflict
+        # 2) Pre-filter: Find which key combinations already exist
+        cursor.execute(
+            f'CREATE TEMP TABLE "{existing_keys_table}" AS '  # noqa: S608
+            f'SELECT {pk_col_list} FROM "{table}" AS r '
+            f'WHERE EXISTS (SELECT 1 FROM "{temp_table}" AS t WHERE {" AND ".join(f'r."{c}" = t."{c}"' for c in pks)})'
         )
-        cursor.execute(insert_sql)
 
-        # 3) Optionally retrieve IDs
+        # 3) Split data into NEW and EXISTING sets
+        if mode == "RAISE":
+            # For RAISE mode, only insert truly new records
+            insert_sql = (
+                f'INSERT INTO "{table}" ({non_id_col_list}) '  # noqa: S608
+                f'SELECT {non_id_col_list} FROM "{temp_table}" AS t '
+                f'WHERE NOT EXISTS (SELECT 1 FROM "{existing_keys_table}" AS e WHERE {" AND ".join(f't."{c}" = e."{c}"' for c in pks)})'
+            )
+            cursor.execute(insert_sql)
+
+        elif mode == "IGNORE":
+            # For IGNORE mode, only insert new records (skip existing)
+            insert_sql = (
+                f'INSERT INTO "{table}" ({non_id_col_list}) '  # noqa: S608
+                f'SELECT {non_id_col_list} FROM "{temp_table}" AS t '
+                f'WHERE NOT EXISTS (SELECT 1 FROM "{existing_keys_table}" AS e WHERE {" AND ".join(f't."{c}" = e."{c}"' for c in pks)})'
+            )
+            cursor.execute(insert_sql)
+
+        else:  # UPDATE mode
+            # First insert new records
+            insert_sql = (
+                f'INSERT INTO "{table}" ({non_id_col_list}) '  # noqa: S608
+                f'SELECT {non_id_col_list} FROM "{temp_table}" AS t '
+                f'WHERE NOT EXISTS (SELECT 1 FROM "{existing_keys_table}" AS e WHERE {" AND ".join(f't."{c}" = e."{c}"' for c in pks)})'
+            )
+            cursor.execute(insert_sql)
+
+            # Then update existing records (only if there are non-pk columns to update)
+            non_pk_cols = [c for c in cols if c not in pks]
+            if non_pk_cols:
+                set_clause = ", ".join(f'"{c}" = t."{c}"' for c in non_pk_cols)
+                update_sql = (
+                    f'UPDATE "{table}" AS r SET {set_clause} '  # noqa: S608
+                    f'FROM "{temp_table}" AS t '
+                    f"WHERE {' AND '.join(f'r."{c}" = t."{c}"' for c in pks)} "
+                    f'AND EXISTS (SELECT 1 FROM "{existing_keys_table}" AS e WHERE {" AND ".join(f't."{c}" = e."{c}"' for c in pks)})'
+                )
+                cursor.execute(update_sql)
+
+        # 4) Optionally retrieve IDs
         id_df = None
         if returning_id:
             buf_out = io.StringIO()
             # Build SELECT clause with explicit aliases
             select_list = ", ".join(['r."ID" AS "ID"', *[f't."{c}" AS "{c}"' for c in pks]])
-            copy_sql = f'COPY (SELECT {select_list} FROM "{table}" AS r JOIN "{temp_table}" AS t USING ({conflict_cols})) TO STDOUT WITH CSV HEADER'  # noqa: S608
+            copy_sql = f'COPY (SELECT {select_list} FROM "{table}" AS r JOIN "{temp_table}" AS t USING ({pk_col_list})) TO STDOUT WITH CSV HEADER'  # noqa: S608
             cursor.copy_expert(copy_sql, buf_out)
             buf_out.seek(0)
             id_df = pl.read_csv(buf_out)
 
-        # 4) Clean up temp table
+        # 5) Clean up temp tables
         cursor.execute(f'DROP TABLE IF EXISTS "{temp_table}"')
+        cursor.execute(f'DROP TABLE IF EXISTS "{existing_keys_table}"')
 
         # Commit this worker's transaction
         raw_conn.commit()
