@@ -1,12 +1,8 @@
 import logging
-import time
-import typing as t
-from datetime import date, datetime
 
 import pandas as pd
-import polars as pl
 
-from ..constants import FREQUENCY_MAPPING, DatasetTypes, FlagLevels
+from ..constants import FlagLevels
 from ..pandas_columns import *
 from .base import encode_boolean_values, validate_dataframe_columns
 
@@ -20,7 +16,7 @@ logger = logging.getLogger(__name__)
 ########################################################################################################
 
 
-def check_for_duplicate_combinations(df: pd.DataFrame, unique_cols: t.List[str], df_name: str = "DataFrame") -> None:
+def check_for_duplicate_combinations(df: pd.DataFrame, unique_cols: list[str], df_name: str = "DataFrame") -> None:
     """
     Check for duplicate combinations in the specified columns of a DataFrame.
 
@@ -31,6 +27,7 @@ def check_for_duplicate_combinations(df: pd.DataFrame, unique_cols: t.List[str],
 
     Raises:
         ValueError: If duplicate combinations are found
+
     """
     if df.duplicated(subset=unique_cols).any():
         num_duplicates = df.duplicated(subset=unique_cols).sum()
@@ -84,9 +81,8 @@ def clean_and_check_store_region_map(store_region_map: pd.DataFrame, copy: bool
         for column in store_region_map.columns:
             if column in ["store", "country"]:
                 continue
-            else:
-                region_type = column
-                break
+            region_type = column
+            break
                 # check that the region type is supported
         if region_type not in StoresColumns.ADDITIONAL_COLUMNS.value:
             raise ValueError(
@@ -131,17 +127,17 @@ def clean_and_check_categories_level_description(
     return categories_level_description
 
 
-def clean_and_check_categories_dict(categories_dict: t.Dict, copy=False) -> t.Dict:
+def clean_and_check_categories_dict(categories_dict: dict, copy=False) -> dict:
     if copy:
         categories_dict = categories_dict.copy()
 
     categories_dict = {int(key): value for key, value in categories_dict.items()}
 
     # Check if the keys are a range from 0 to n-1 without gaps
-    if set(categories_dict.keys()) != set(range(0, len(categories_dict))):
+    if set(categories_dict.keys()) != set(range(len(categories_dict))):
         keys = sorted(categories_dict.keys())
         raise ValueError(
-            "The keys of the categories dictionary must be a range from 0 to n-1 without gaps, got: {}".format(keys)
+            f"The keys of the categories dictionary must be a range from 0 to n-1 without gaps, got: {keys}"
         )
 
     return categories_dict
@@ -382,9 +378,8 @@ def clean_and_check_time_region_feature_data(
         for column in time_region_feature_data.columns:
             if column in ["date", "country", "feature", "value"]:
                 continue
-            else:
-                region_type = column
-                break
+            region_type = column
+            break
                 # check that the region type is supported
         if region_type not in TimeRegionFeatureColumns.ADDITIONAL_COLUMNS.value:
             raise ValueError(
